# Grid Discretization Helper Copyright (c) 2016, The Regents of the University
# of California, through Lawrence Berkeley National Laboratory (subject to
# receipt of any required approvals from the U.S. Dept. of Energy).  All rights
# reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Innovation & Partnerships Office at:
# IPO@lbl.gov.
#
# NOTICE.  This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights. As
# such, the U.S. Government has been granted for itself and others acting on
# its behalf a paid-up, non-exclusive, irrevocable, worldwide license in the
# Software to reproduce, distribute copies to the public, prepare derivative
# works, and perform publicly and display publicly, and to permit others to do
# so.

import pygdh
import numpy

# GDB needs 'Grid' objects for storing solution results, even though this
# problem is not position-dependent
class FalseDomain(pygdh.Grid):
    
    def __init__(self,name):

        # This 'Grid' object only contains a single 'Volume' object
        self.initialize_grid(name, field_names=['y'],
                             # As there is no spatial dependence, it is not
                             # necessary to declare independent spatial
                             # variables
                             unit_classes=[pygdh.Volume])

    def set_initial_conditions(self):

        # Set the value of the single dependent variable directly
        self.field[0][0,0] = 1.0

class ODE_in_time(pygdh.Problem):

    def __init__(self, alpha):

        # Save the argument value as an object member
        self.alpha = alpha

        grid_obj = FalseDomain('false_domain')

        # Initialize data structures, among other things
        self.initialize_problem([grid_obj],past_timestep_count=1)

    def ode_in_time(self, vol, residual):

        # There is only one solution field, with index 0, and with one
        # ``Volume``, with label 0        
        y = self.grid[0].field[0][0,0]
        y_1 = self.grid[0].field[-1][0,0]

        residual[0] = (y - y_1)*self.inverse_timestep_size + self.alpha*y

    def set_equation_scalar_counts(self):

        self.equation_scalar_counts = {self.ode_in_time: 1}
        
    def assign_equations(self):

        self.equations[0][0] = [self.ode_in_time]

# Create an object instance, setting alpha = 1.0
problem = ODE_in_time(1.0)

number_of_timesteps = 100
timestep_size = 0.01

# The names of all output files will begin with this string
filename_root = 'ODE_in_time'

# One output file will be created for each format given here
output_types = ['GNUPLOT']

# Calculate solution and write output to file
problem.solve_time_dependent_system(number_of_timesteps*timestep_size,
                                    timestep_size, filename_root, output_types)

