# Grid Discretization Helper Copyright (c) 2016, The Regents of the University
# of California, through Lawrence Berkeley National Laboratory (subject to
# receipt of any required approvals from the U.S. Dept. of Energy).  All rights
# reserved.
#
# If you have questions about your rights to use or distribute this software,
# please contact Berkeley Lab's Innovation & Partnerships Office at:
# IPO@lbl.gov.
#
# NOTICE.  This Software was developed under funding from the U.S. Department
# of Energy and the U.S. Government consequently retains certain rights. As
# such, the U.S. Government has been granted for itself and others acting on
# its behalf a paid-up, non-exclusive, irrevocable, worldwide license in the
# Software to reproduce, distribute copies to the public, prepare derivative
# works, and perform publicly and display publicly, and to permit others to do
# so.

import pygdh
import numpy
import math

# Objects of this class are the units from which the spatial domain will be
# constructed.
class Volume(pygdh.Volume):

    # Defining this method gives users an opportunity to define their own
    # mathematical operators, for later use in describing governing equations.
    def define_interpolants(self):

        # Operator for estimating first derivative at left boundary of volume
        self.dx_left = self.default_interpolant(-1.0, 1, 1)
        
        # Operator for estimating function value at the volume center
        self.interpolate = self.default_interpolant(0.0, 0, 2)
        
        # Operator for estimating first derivative at right boundary of volume
        self.dx_right = self.default_interpolant(1.0, 1, 1)

# Objects of this class describe the spatial domains on which problems will be
# solved.
class Domain(pygdh.Grid):

    # This method is automatically called immediately after an object of this
    # class is created. Users are free to modify this method definition as
    # desired, as long as the 'self.initialize_grid' method is called with
    # appropriate parameter values at some point in this method.
    def __init__(self, name, alpha, unit_count, x_coordinates):

        # Save the argument value to an object member
        self.alpha = alpha

        self.initialize_grid(name,
                             # The independent variable will be named 'y'
                             field_names=['y'],
                             # The dependent variable will be named 'x',
                             # and 'y' will be computed at the positions
                             # specified by the x_coordinates parameter
                             coordinate_values={'x': x_coordinates},
                             # Each unit in 'Domain' will be described by a
                             # 'Volume' object                             
                             unit_classes=[Volume for i in range(unit_count)])

    # This sets the initial values of the independent variables on the region
    # represented by an object of this class.
    def set_initial_conditions(self):

        # Defined for clarity
        y = self.field[0][0]

        # Set the value of the independent variable on each volume
        for vol in self.unit_with_number:
            y[vol.number] = (
                vol.coordinate / self.alpha
                - 0.5*math.sin(0.5*math.pi*vol.coordinate))

    # This method is required. It informs GDB of the unknown quantities for
    # for which it must solve.
    def declare_unknowns(self):

        # left boundary
        vol = self.unit_with_number[0]
        vol.unknown_field[0] = False

        # Interior volumes
        for vol_number in range(1, self.unit_count-1):
            vol = self.unit_with_number[vol_number]
            vol.unknown_field[0] = True

        # right boundary
        vol = self.unit_with_number[-1]
        vol.unknown_field[0] = True

# Objects of this class bring the spatial domain definitions together with
# equations, boundary conditions, and for time-dependent problems, initial
# conditions. These objects can then be directed to compute numerical
# solutions.
class MultipleGrids(pygdh.Problem):

    def __init__(self, alpha1, unit_count1, alpha2, unit_count2):

        # Save the argument value as an object member
        self.alpha1 = alpha1
        self.alpha2 = alpha2

        # The left and right domains are both objects of the 'Domain' class,
        # but they must be initialized to represent different regions
        grid_obj1 = Domain('left_domain',alpha1,unit_count1,
                           numpy.linspace(0., 0.5, unit_count1))
        grid_obj2 = Domain('right_domain',alpha2,unit_count2,
                           numpy.linspace(0.5, 1., unit_count2))

        # Prepare to run simulation using two Grid objects, numbered in this
        # order
        self.initialize_problem([grid_obj1,grid_obj2],past_timestep_count=1)
        
    def left_grid_pde(self, vol, residuals):

        # Present solution values
        y0 = self.grid[0].field[0][0]

        # Past solution values
        y_1 = self.grid[0].field[-1][0]

        residuals[0] = ((vol.interpolate(y0) - vol.interpolate(y_1))*vol.volume
                        - self.timestep_size * self.alpha1
                        * (vol.dx_right(y0) - vol.dx_left(y0)))

    def left_grid_center_pde(self, vol, residuals):

        # Present solution values
        y0 = self.grid[0].field[0][0]

        # Past solution values
        y_1 = self.grid[0].field[-1][0]

        right_y0 = self.grid[1].field[0][0]
        right_vol = self.grid[1].unit_with_number[0]

        residuals[0] = ((vol.interpolate(y0) - vol.interpolate(y_1))*vol.volume
                        - self.timestep_size * self.alpha1
                        # The derivative at the interface will be obtained by
                        # averaging the one-sided differences taken from each
                        # side
                        * (0.5*(vol.dx_right(y0) + right_vol.dx_left(right_y0))
                           - vol.dx_left(y0)))

    def right_grid_pde(self, vol, residuals):

        # Present solution values
        y0 = self.grid[1].field[0][0]

        # Past solution values
        y_1 = self.grid[1].field[-1][0]

        residuals[0] = ((vol.interpolate(y0) - vol.interpolate(y_1))*vol.volume
                        - self.timestep_size * self.alpha2
                        * (vol.dx_right(y0) - vol.dx_left(y0)))

    def right_grid_right_boundary_pde(self, vol, residuals):

        # Present solution values
        y0 = self.grid[1].field[0][0]

        # Past solution values
        y_1 = self.grid[1].field[-1][0]

        residuals[0] = ((vol.interpolate(y0) - vol.interpolate(y_1))*vol.volume
                        + self.timestep_size
                        * (-1.0 + self.alpha2 * vol.dx_left(y0)))

    def set_boundary_values(self):

        ## Boundary with the prescribed value is on the left side of the left
        ## domain
        y_left = self.grid[0].field[0][0]

        # Left boundary
        y_left[0] = 0.

        ## In the right domain, the value at left boundary is taken from the
        ## adjacent left domain
        y_right = self.grid[1].field[0][0]

        # Left boundary
        y_right[0] = y_left[-1]

    def set_equation_scalar_counts(self):

        # This is a dictionary associating governing equation methods with
        # the number of scalar values that they return
        self.equation_scalar_counts = {self.left_grid_pde: 1,
                                       self.left_grid_center_pde: 1,
                                       self.right_grid_pde: 1,
                                       self.right_grid_right_boundary_pde: 1}

    def assign_equations(self):

        ## Left grid

        left_grid_equations = self.equations[0]

        # Left boundary
        left_grid_equations[0] = []

        # Interior volumes
        for vol_number in range(1, self.grid[0].unit_count-1):
            left_grid_equations[vol_number] = [self.left_grid_pde]

        # Right boundary
        left_grid_equations[-1] = [self.left_grid_center_pde]
        
        ## Right grid

        right_grid_equations = self.equations[1]

        # Left boundary, value to be copied from left grid
        right_grid_equations[0] = []

        # Interior volumes
        for vol_number in range(1, self.grid[0].unit_count-1):
            right_grid_equations[vol_number] = [self.right_grid_pde]

        # Right boundary
        right_grid_equations[-1] = [self.right_grid_right_boundary_pde]

# Create an object instance, setting alpha1 = 1.0, alpha2 = 1.0
problem = MultipleGrids(1.0, 6, 1.0, 6)

number_of_timesteps = 3
timestep_size = 0.01

# The names of all output files will begin with this string
filename_root = 'multiple_grids'

# One output file will be created for each format given here
output_types = ['GNUPLOT']

# Calculate solution and write output to file
problem.solve_time_dependent_system(number_of_timesteps*timestep_size,
                                    timestep_size, filename_root,
                                    output_types)

